#!/usr/bin/env python

# AUTHOR:   Adam Wehmann
# DATE:     12/16/2012
# WEBSITE:  http://www.adamwehmann.com/
#
# PURPOSE:  Creates time-geographic object in shape of prism
#           from a set of points with a given time budget (in
#           terms of time T1 and T2) and travel velocity given
#           as attributes, illustrating the potential path
#           space of the point entity.  Also created is the
#           'shadow' of the prism, which corresponds to the
#           potential path area of the entity.

from math import tan
from math import cos
from math import sin
from math import pi
import arcpy
import os

# PARAMETERS --------------------------------------------------

# Point shapefile containing sites at which to draw prisms.
#   Must have fields T1, T2, and V corresponding to time and
#   velocity values to define prism.
filein = "C:\\GISData\\cone\\points.shp"

# Output shapefile to hold 3D polygon data.
#   'Id' field will match original point layer's 'FID' field.
fileout = "C:\\GISData\\cone\\test2.shp"

# Output shapefile to hold prism shadows.
projout = "C:\\GISData\\cone\\test_proj2.shp"

# Vertical scaling factor
sfactor = 10.0

# Number of polygons to use to generate each half of each
# prism.  A higher number will ensure prism cross sections
# appear circular.
nfidelity = 100

# Whether to allow outputs to be overwritten
overwrite = True

# FUNCTIONS --------------------------------------------------

# CREATES SEMICIRCLE AT (x,y) WITH RADIUS r
def semicircle1_poly(x,y,r,n=50): 
    for i in range(n+1):
        ang = pi * (n-i)/n
        yield (x + r * cos(ang), y + r * sin(ang) )

def semicircle2_poly(x,y,r,n=50): 
    for i in range(n+1):
        ang = -pi * (n-i)/n
        yield (x + r * cos(ang), y + r * sin(ang) )

# WRITES PRISM TO SHAPEFILE
# INPUT:     x - x-position
#            y - y-position
#           t1 - time 1
#           t2 - time 2
#            v - travel velocity
#            s - (optional) scaling factor
#			 n - (optional) number of polygon parts per prism half
#           sr - (optional) spatial reference
# OUTPUT: tuple (xyz_t, xyz_b)
#    where xyz_t - list of polygons in xyz-domain defining top of prism
#          xyz_b - list of polygons in xyz-domain defining bottom of prism
def create_timeprism(x,y,t1,t2,v,s=1.0,n=100,sr=None):
    # Get radius at widest cross-section.
    r = s*(t2-t1)/tan(1.0/v)
    # Get height at widest cross-section
    z = s*(t1 + (t2-t1)/2)
    # Create prism
    th1 = []
    bh1 = []
    first = True
    for (px,py) in semicircle1_poly(x,y,r,n/2):
        if not first:
            # Top half of upper half of prism
            sarr1 = arcpy.Array([arcpy.Point(x,y,s*t2),arcpy.Point(lpx,lpy,z),arcpy.Point(px,py,z)])
            sarr1.add(arcpy.Point(x,y,s*t2))
            th1.append(arcpy.Polygon(sarr1,sr,True))
            # Top half of bottom half of prism
            sarr2 = arcpy.Array([arcpy.Point(x,y,s*t1),arcpy.Point(lpx,lpy,z),arcpy.Point(px,py,z)])
            sarr2.add(arcpy.Point(x,y,s*t1))
            bh1.append(arcpy.Polygon(sarr2,sr,True))
        else:
            first = False
        lpx = px
        lpy = py
    th2 = []
    bh2 = []
    first = True
    for (px,py) in semicircle2_poly(x,y,r,n/2):
        if not first:
            # Bottom half of upper half of prism
            sarr1 = arcpy.Array([arcpy.Point(x,y,s*t2),arcpy.Point(px,py,z),arcpy.Point(lpx,lpy,z)])
            sarr1.add(arcpy.Point(x,y,s*t2))
            th2.append(arcpy.Polygon(sarr1,sr,True))
            # Bottom half of bottom half of prism
            sarr2 = arcpy.Array([arcpy.Point(x,y,s*t1),arcpy.Point(px,py,z),arcpy.Point(lpx,lpy,z)])
            sarr2.add(arcpy.Point(x,y,s*t1))
            bh2.append(arcpy.Polygon(sarr2,sr,True))
        else:
            first = False
        lpx = px
        lpy = py
    # Create feature list of polygons defining upper half of prism
    xyz_t = []
    xyz_t.extend(th1)
    xyz_t.extend(th2)
    # Create feature list of polygons defining bottom half of prism
    xyz_b = []
    xyz_b.extend(bh1)
    xyz_b.extend(bh2)
    # Return results
    return (xyz_b, xyz_t)

# MAIN FILE --------------------------------------------------  

# Describe
desc = arcpy.Describe(filein)
sr = desc.spatialReference

# Check fields exist
c = 0
for field in desc.fields:
    if field.name == "T1" or field.name == "T2" or field.name == "V":
        c += 1
if c < 3:
    print "Input point shapefile must have fields named T1, T2, V."
    exit()

# Identify the geometry field
shapefieldname = desc.ShapeFieldName
# Identify the OID field
oidfieldname = desc.OIDFieldName

# Create search cursor
rows = arcpy.SearchCursor(filein)

# Empty feature list
f = []
l = []

# Loop through features
for row in rows:
    
    # Create geometry object for feature
    feat = row.getValue(shapefieldname)
    pnt = feat.getPart()
    
    # Get attribute values
    oid = row.getValue(oidfieldname)
    t1 = row.getValue('T1')
    t2 = row.getValue('T2')
    v = row.getValue('V')

    # Create prism
    (xyz_t,xyz_b) = create_timeprism(pnt.X,pnt.Y,t1,t2,v,s=sfactor,n=nfidelity,sr=sr)
    
    # Add new geometry
    f.extend(xyz_t)
    f.extend(xyz_b)

    # Maintain list of OIDs corresponding to features
    l.extend([oid]*(len(xyz_t)+len(xyz_b)))

del rows, row

# Output
if fileout != '':
    if os.path.exists(fileout) and overwrite:
        arcpy.DeleteFeatures_management(fileout)
    arcpy.CreateFeatureclass_management(os.path.dirname(fileout),os.path.basename(fileout), "POLYGON", None, "DISABLED", "ENABLED", sr)
    cur = arcpy.InsertCursor(fileout)
    for (x,y) in zip(f,l):
        feat = cur.newRow()
        feat.Id = y
        feat.shape = x
        cur.insertRow(feat)
    del cur, feat

    if projout != '':
        if os.path.exists(projout) and overwrite:
            arcpy.DeleteFeatures_management(fileout)
        arcpy.Dissolve_management(fileout, projout,["Id"])

# All done
print "Finished!"
